use gcd::Gcd;

/**
 * Calculate the primedivisor for some number `n`
*/
pub fn pollard_rho(n: u16) -> u16 {
    if n == 1 { return 1; } // 1 only has 1 as prime divisor
    if n % 2 == 0 { return 2; } // even numbers have at least 2 as prime divisor
    if is_prime(n) { return n; }

    let mut x = rand::random::<u16>() % 2;
    let mut y = x;

    let c = rand::random::<u16>() % 2;
    let mut div = 1;

    while div == 1 {
        x = (mod_pow(x, 2, n) + c + n) % n;
        y = (mod_pow(y, 2, n) + c + n) % n;
        y = (mod_pow(y, 2, n) + c + n) % n;
        div = u16::try_from((i16::try_from(x).unwrap()-i16::try_from(y).unwrap()).abs()).unwrap().gcd(n);
        eprintln!("Got div {}", div);
        if div == n { return pollard_rho(n); }
    }

    return div;
}

/**
 * Discrete/Modular exponentiation
 * 
 * Highly memory efficient because the full result is never stored, but shortened by defined modulo instead. 
 * We can use that because the prime divisor required for our algorithm is guarenteed to be smaller
 * than n.
 * 
 * Counterpart function to the discrete logarithm.
 */
fn mod_pow(base: u16, exp: u16, r#mod: u16) -> u16 {
    let mut result = 1; 

    for _ in 0..exp-1 {
        result = (result * base) % r#mod;
    }

    result
}

fn is_prime(n: u16) -> bool {
    for i in (3..=(n as f32).sqrt() as u16).step_by(2) {
        if n % i == 0 {  return false; }
    }
    true
}
